#! /usr/bin/perl

# Script converts Scaleble Vector Graphics int Gcode commands for plotting
# More infos on the project page:
# https://homofaciens.de/technics-machines-3D-printer-Zonestar-Plotter_en.htm

use strict;
use IO::Handle;

my $i = 0;
my $j = 0;
my $line_in = "";

my $progress = 0;
my $start = 0;
my $stop = 0;
my $X=0;
my $Y=0;
my $Z=0;
my $feed_rate = 2.000;
my $feed_rate_empty = (100.000 * 60);#feedrate in mm per minute
my $feed_rate_material = (30.00 * 60);#feedrate in mm per minute
my $feed_rate_pen_change = (50.00 * 60);#feedrate in mm per minute

my $Z_up = 10.100;
my $Z_down = 0.501;#Z=0.000 would trigger Z switch whenever Pen is down

my @feedrate_array;
my @remarks;
my @X_coords;
my @Y_coords;
my @Z_coords;

my $coord_no = 0;


my $x_min = 99999;
my $x_max = -99999;
my $y_min = 99999;
my $y_max = -99999;

my $bed_size_x = 220;
my $bed_size_y = 220;
my $x_offset = 0;#Offset needed to center plot
my $y_offset = 0;#Offset needed to center plot

#Pen position correction relative to printer nozzle
my $pen_Y = 0;
my $pen_X = 0;

my $scale = 1.00;
my $mirror_X = 0;
my $mirror_Y = 0;

my $SourceFile = "";

my $rgb_r = 0;
my $rgb_g = 0;
my $rgb_b = 0;

my $color = 0;
my $color_new = -1;
my $color_old = -1;

my @pen_X;
my @pen_gap_X;
my @pen_Z_low;
my @pen_Z_low_empty;
my @pen_Z_high;

#Set $color_no = 1 for monochrome mode
my $color_no = 4;

#Coordinates pen 1
push @pen_X, -12;
push @pen_gap_X, 33;
push @pen_Z_low, 30;
push @pen_Z_low_empty, 79;
push @pen_Z_high, 190;

#Coordinates pen 2
push @pen_X, 77;
push @pen_gap_X, 33;
push @pen_Z_low, 30;
push @pen_Z_low_empty, 79;
push @pen_Z_high, 190;

#Coordinates pen 3
push @pen_X, 127;
push @pen_gap_X, 171;
push @pen_Z_low, 30;
push @pen_Z_low_empty, 79;
push @pen_Z_high, 190;

#Coordinates pen 4
push @pen_X, 214;
push @pen_gap_X, 171;
push @pen_Z_low, 30;
push @pen_Z_low_empty, 79;
push @pen_Z_high, 190;



my $Graphic = "";

my $FileMode = 1;

my $SourceFile = "";

my $num_args = $#ARGV + 1;

my $PI = 3.1415927;

if($num_args < 1){
  print "Call without parameters, exiting.\n";
  exit(1);
}
else{
  for($i=0; $i<$num_args; $i++){
    if(index($ARGV[$i], ".svg")>-1){
      $SourceFile = substr($ARGV[$i], 0, index($ARGV[$i], ".svg"));
    }
    print"Parameter number $i = $ARGV[$i]\n";
    if(index($ARGV[$i], "CIRCLES")>-1){
      $Graphic = "CIRCLES";
    }
    if(index($ARGV[$i], "SPIRALS")>-1){
      $Graphic = "SPIRALS";
    }
    if(index($ARGV[$i], "FUNCTIONS")>-1){
      $Graphic = "FUNCTIONS";
    }
    if(index($ARGV[$i], "SQUARES")>-1){
      $Graphic = "SQUARES";
    }
  }
}

if($Graphic eq "SQUARES"){
  my $r = 10;
  my $alpha = 0;
  my $arrow = 5;
  my $factor = 0.01;
  my $Yoffset = 0;
  my $line_l = 0;
  my $line_start = 5;
  my $line_plus = 3;
  my $square_size = 65;
  my $gap = 2;


  for($i=0; $i<4; $i++){
    if($i == 0){
      $color = 1;
      &PickPen($color);
      $X=-$line_start + $square_size/2 + $gap;
      $Y=$line_start + $square_size/2 + $gap;
    }
    if($i == 1){
      $X=-$line_start - $square_size/2 - $gap;
      $Y=$line_start - $square_size/2 - $gap;
    }
    if($i == 2){
      $color = 2;
      &PickPen($color);
      $X=-$line_start - $square_size/2 - $gap;
      $Y=$line_start + $square_size/2 + $gap;
    }
    if($i == 3){
      $X=-$line_start + $square_size/2 + $gap;
      $Y=$line_start - $square_size/2 - $gap;
    }

    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    &AddCoordinates();
    $Z = $Z_down;
    $feed_rate = $feed_rate_material;
    &AddCoordinates();

    for($line_l = $line_start; $line_l < $square_size; $line_l += $line_plus){
      $X += $line_l;
      &AddCoordinates();
      $line_l += $line_plus;
      $Y += $line_l;
      &AddCoordinates();
      $line_l += $line_plus;
      $X -= $line_l;
      &AddCoordinates();
      $line_l += $line_plus;
      $Y -= $line_l;
      &AddCoordinates();
    }
    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    &AddCoordinates();
    if($i==1 || $i == 3){
      &ParkPen($color);
    }
  }



  $FileMode = 0;
  $SourceFile = "SQUARES";
}


if($Graphic eq "FUNCTIONS"){
  my $r = 10;
  my $alpha = 0;
  my $arrow = 5;
  my $factor = 0.01;
  my $Yoffset = 0;
  my $coord_size_x = 200;
  my $coord_size_y = 200;

  $mirror_X = 0;
  $mirror_Y = 1;

  &PickPen(0);


  #Y axis
  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  $X = 0;
  $Y = -$coord_size_y/2;
  &AddCoordinates();
  $Z = $Z_down;
  $feed_rate = $feed_rate_material;
  &AddCoordinates();
  $Y = $coord_size_y/2;
  &AddCoordinates();
  $X = $arrow;
  $Y = $coord_size_y/2-$arrow;
  &AddCoordinates();
  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  &AddCoordinates();
  $X = 0;
  $Y = $coord_size_y/2;
  &AddCoordinates();
  $Z = $Z_down;
  $feed_rate = $feed_rate_material;
  &AddCoordinates();
  $X = -$arrow;
  $Y = $coord_size_y/2-$arrow;
  &AddCoordinates();
  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  &AddCoordinates();

  for($Y=-90; $Y<=90; $Y+=10){
    $X = -5;
    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    &AddCoordinates();
    $Z = $Z_down;
    $feed_rate = $feed_rate_material;
    &AddCoordinates();
    $X = 5;
    &AddCoordinates();
    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    &AddCoordinates();
  }

  #X axis
  $X = -$coord_size_x/2;
  $Y = 0;
  &AddCoordinates();
  $Z = $Z_down;
  $feed_rate = $feed_rate_material;
  &AddCoordinates();
  $X = $coord_size_x/2;
  &AddCoordinates();
  $Y = $arrow;
  $X = $coord_size_x/2-$arrow;
  &AddCoordinates();
  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  &AddCoordinates();
  $Y = 0;
  $X = $coord_size_x/2;
  &AddCoordinates();
  $Z = $Z_down;
  $feed_rate = $feed_rate_material;
  &AddCoordinates();
  $Y = -$arrow;
  $X = $coord_size_x/2-$arrow;
  &AddCoordinates();
  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  &AddCoordinates();
  for($X=-90; $X<=90; $X+=10){
    $Y = -5;
    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    &AddCoordinates();
    $Z = $Z_down;
    $feed_rate = $feed_rate_material;
    &AddCoordinates();
    $Y = 5;
    &AddCoordinates();
    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    &AddCoordinates();
  }

  &ParkPen(0);



  for($color=1; $color<4; $color++){
    &PickPen($color);
    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    if($color == 1){
      $factor = 0.02;
      $Yoffset = 100;
    }
    if($color == 2){
      $factor = 0.015;
      $Yoffset = 80;
    }
    if($color == 3){
      $factor = 0.003;
      $Yoffset = 0;
    }
    for($X = -$coord_size_x; $X<$coord_size_x; $X+=0.5){
      $Y = $factor*$X * $X - $Yoffset;
      if($X<$coord_size_x/2 && $X>-$coord_size_x/2 && $Y>-$coord_size_y/2 && $Y<$coord_size_y/2){
        &AddCoordinates();
        if($Z == $Z_up){
          $Z = $Z_down;
          $feed_rate = $feed_rate_material;
        }
        &AddCoordinates();
      }
    }
    &ParkPen($color);
  }



  $FileMode = 0;
  $SourceFile = "FUNCTIONS";
}

if($Graphic eq "SPIRALS"){
  my $r = 10;
  my $alpha = 0;


  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  $X = 0;
  $Y = 0;
  &AddCoordinates();
  $Z = $Z_down;
  $feed_rate = $feed_rate_material;


  for($color=0; $color<2; $color++){
    $Z = $Z_up;
    $feed_rate = $feed_rate_empty;
    $X = 0;
    $Y = 0;
    &AddCoordinates();
    &PickPen($color);
    $X = 0;
    $Y = 0;
    &AddCoordinates();
    $Z = $Z_down;
    $feed_rate = $feed_rate_material;
    $r += 1;
    for($alpha = 0.0; $alpha<$PI*2*10; $alpha+=0.01){
      $r = $alpha * 1.5+$color*3;
      $X = cos($alpha) * $r;
      $Y = sin($alpha) * $r;
      &AddCoordinates();
    }
    &ParkPen($color);
  }



  $FileMode = 0;
  $SourceFile = "SPIRALS";
}

if($Graphic eq "CIRCLES"){
  my $ry = 10;
  my $rx = 50;
  my $alpha = 0;

  &PickPen(3);

  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  $X = cos($alpha) * $rx+1;
  $Y = sin($alpha) * $ry;
  &AddCoordinates();
  $Z = $Z_down;
  $feed_rate = $feed_rate_material;


  for($ry=10; $ry<100; $ry+=10){
    $rx += 1;
    for($alpha = 0.0; $alpha<$PI*2; $alpha+=0.01){
      $X = cos($alpha) * $rx;
      $Y = sin($alpha) * $ry;
      &AddCoordinates();
    }
  }

  &ParkPen(3);

#  open( COM, "/dev/ttyUSB0") || die "Cannot read serial port : $!\n";
#  while( <COM> ){
#      # do stuff here
#  }

  $FileMode = 0;
  $SourceFile = "CIRCLES";
}


if($FileMode == 1){
  for($color = 0; $color < $color_no; $color++){
    open(FILEIN,'< ' . $SourceFile . '.svg') || die "Can't open $SourceFile.svg!\n";

      $x_min = 99999;
      $x_max = -99999;
      $y_min = 99999;
      $y_max = -99999;


      while(!(eof(FILEIN))){
        $line_in = <FILEIN>;
        $line_in =~ s/^\s+|\s+$//g;

        #check for color
        if($color_no > 1){
          if(index($line_in, "rgb(") > -1){
            my $line_temp = $line_in;
            $start = index($line_temp, "rgb(");
            $start += 4;
            $line_temp = substr($line_temp, $start);
            $stop = index($line_temp, ",");
            $rgb_r = substr($line_temp, 0, $stop);
            $start = $stop + 1;
            $line_temp = substr($line_temp, $start);
            $stop = index($line_temp, ",");
            $rgb_g = substr($line_temp, 0, $stop);
            $start = $stop + 1;
            $line_temp = substr($line_temp, $start);
            $stop = index($line_temp, ")");
            $rgb_b = substr($line_temp, 0, $stop);

            if($rgb_r < 128){
              $rgb_r = 0;
            }
            else{
              $rgb_r = 255;
            }

            if($rgb_g < 128){
              $rgb_g = 0;
            }
            else{
              $rgb_g = 255;
            }

            if($rgb_b < 128){
              $rgb_b = 0;
            }
            else{
              $rgb_b = 255;
            }

            $color_new = 0;#Pen 0 is default color
            if($rgb_r == 0 && $rgb_g == 0 && $rgb_b == 0){
              $color_new = 0;
            }

            if($rgb_r == 255 && $rgb_g == 0 && $rgb_b == 0){
              $color_new = 1;
            }

            if($rgb_r == 0 && $rgb_g == 255 && $rgb_b == 0){
              $color_new = 2;
            }

            if($rgb_r == 0 && $rgb_g == 0 && $rgb_b == 255){
              $color_new = 3;
            }
            #White color, pen 99 is ignored (0-3)!
            if($rgb_r == 255 && $rgb_g == 255 && $rgb_b == 255){
              $color_new = 99;
            }

            print"R=$rgb_r, G=$rgb_g, B=$rgb_b, color_new=$color_new\n";
          }
          if($color_new != $color){
            $line_in = "";#Ignore line
          }
          else{
            if($color_old != $color){#Change pen
              if($color_old != -1){#Start plotting, no need to park a pen
                print"color_old=$color_old, color=$color\n";
                &ParkPen($color_old);
              }
              &PickPen($color);
              $color_old = $color;
            }
          }
        }
        if(index($line_in, "<path fill=\"none\"") > -1){
          $start = index($line_in, "d=\"M ");
          if($start > -1){
            $X = 0;
            $Y = 0;
            $Z = $Z_up;
            $feed_rate = $feed_rate_empty;
            $start += 5;
            $line_in = substr($line_in, $start);
            $stop = index($line_in, ",");
            $X = substr($line_in, 0, $stop) * 0.01 * $scale;

            $start = $stop + 1;
            $line_in = substr($line_in, $start);
            $stop = index($line_in, " ");
            $Y = substr($line_in, 0, $stop) * 0.01 * $scale;
            if($X > $x_max){
               $x_max = $X;
            }
            if($X < $x_min){
              $x_min = $X;
            }
            if($Y > $y_max){
              $y_max = $Y;
            }
            if($Y < $y_min){
              $y_min = $Y;
            }

            &AddCoordinates();
            #print FILEOUT "X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . "\n";
            $Z = $Z_down;
            $feed_rate = $feed_rate_material;

            &AddCoordinates();
            #print FILEOUT "X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . "\n";


            $start = $stop + 3;
            $line_in = substr($line_in, $start);
            while($stop > -1){
              #print $line_in . "\n";
              $stop = index($line_in, ",");
              if($stop > -1){
                $X = substr($line_in, 0, $stop) * 0.01 * $scale;
                $start = $stop + 1;
                $line_in = substr($line_in, $start);
                $stop = index($line_in, " ");
                $Y = substr($line_in, 0, $stop) * 0.01 * $scale;
                $start = $stop + 1;
                $line_in = substr($line_in, $start);
                if($X > $x_max){
                  $x_max = $X;
                }
                if($X < $x_min){
                  $x_min = $X;
                }
                if($Y > $y_max){
                  $y_max = $Y;
                }
                if($Y < $y_min){
                  $y_min = $Y;
                }
                &AddCoordinates();
                #print FILEOUT "X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . "\n";
              }
            }
            #print FILEOUT ";UP X$X Y$Y\n";
            $Z = $Z_up;
            $feed_rate = $feed_rate_empty;
            &AddCoordinates();
            #print FILEOUT "X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . "\n";

          }
          else{
            printf "d=\"M  not found!\n";
            $progress = -1;
          }
        }

      }#while(!(eof(FILEIN)))

    close(FILEIN);

  }#for $color
}#if($FileMode == 1);

if($color_old != -1){#finally park pen
  &ParkPen($color_old)
}

print $coord_no . " coordinates found. Processing data...\n";
open(FILEOUT,'> ' . $SourceFile . '.gcode') || die "Can not open $SourceFile.gcode!\n";
  #print FILEOUT "M3 ;Spindle ON"
  #print FILEOUT "M8 ;Coolant ON"

  #needed to center plot on build plate
  $x_offset = ($bed_size_x-($x_max-$x_min))/2;
  $y_offset = ($bed_size_y-($y_max-$y_min))/2;

  print FILEOUT "G21 ; set units to millimeters\n";
  print FILEOUT "G90 ; use absolute coordinates\n";
  print FILEOUT "G1 Z" . sprintf("%.3f", $Z) . " ; Lift pen\n";
  print FILEOUT "G28 X0 Y0 ; home X and Y axis\n";
  print FILEOUT "G28 Z0 ; home Z axis\n";
  print FILEOUT "M203 X200 Y200 Z100 ; Set max feedrates\n";#MARLIN version (M203), ignored by Repetier

  #Start at 0/0 z_up
  $X = 0;
  $Y = 0;
  $Z = $Z_up;
  $X =  $x_offset;
  $Y =  $y_offset;
  $Z =  $Z;

  print FILEOUT "G1 Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . " ; Move Pen to UP position\n";
  print FILEOUT "G1 X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . " ; Move pen to 0/0\n";


  if($x_offset < 0){
    print "Width of file too large for print bed!\n";
  }
  if($y_offset < 0){
    print "Height of file too large for print bed!\n";
  }

  for($i = 0; $i < $coord_no; $i++){
    if(index($remarks[$i], "keep X")>-1){
      $X = $X;#Use previous X coordinate
    }
    else{
      if(index($remarks[$i], "PenAction:")==-1){
        if($mirror_X == 0){
          $X = $X_coords[$i] - $x_min + $x_offset + $pen_X;
        }
        else{
          $X = ($x_max - $X_coords[$i]) + $x_offset + $pen_X;
        }
      }
      else{
        $X = $X_coords[$i];
      }
    }

    if(index($remarks[$i], "keep Y")>-1){
      $Y = $Y;#Use previous Y coordinate
    }
    else{
      if(index($remarks[$i], "PenAction:")==-1){
        if($mirror_Y == 0){
          $Y = ($y_max - $Y_coords[$i]) + $y_offset + $pen_Y;
        }
        else{
          $Y = $Y_coords[$i] - $y_min + $y_offset + $pen_Y;
        }
      }
      else{
        $Y = $Y_coords[$i];
      }
    }

    $Z =  $Z_coords[$i];
    if($Z == $Z_up){
      $feed_rate = $feed_rate_empty;
    }
    else{
      $feed_rate = $feed_rate_material;
    }
    print FILEOUT "G1 X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feedrate_array[$i]);
    if(length($remarks[$i])>0){
      print FILEOUT "; " . $remarks[$i];
    }

    print FILEOUT "\n";
  }

  $X = 0;
  $Y = 0;
  $Z = $Z_up;
  $feed_rate = $feed_rate_empty;
  print FILEOUT "X" . sprintf("%.3f", $X)  . " Y" . sprintf("%.3f", $Y)  .  " Z" . sprintf("%.3f", $Z)  . " F" . sprintf("%.3f", $feed_rate)  . "\n";
  #print FILEOUT "M3\n";
  #print FILEOUT "M8\n";

  #print FILEOUT "M5 ;Spindle OFF"
  #print FILEOUT "M9 ;Coolant OFF"

  print FILEOUT "G1 Z0" . sprintf("%.3f", $Z_up) . " ; Lift pen\n";
  print FILEOUT "G28 X0 ; home X axis\n";
  print FILEOUT "M84 ; disable all stepper motors\n";


  print "x_max = $x_max, x_min = $x_min, ymax = $y_max, y_min = $y_min \n";
  print "width = " . ($x_max - $x_min) . ", height = " . ($y_max- $y_min) . "\n";
  print "X offset = " . ($x_offset) . ", Y offset = " . ($y_offset) . "\n";

  print "...done!\n";

close(FILEOUT);

if(($x_max - $x_min) > $bed_size_x){
  print "\n\n########################################################\n";
  print "Width of plot larger than width of print bed !!!!\n";
  print "Change value of \$scale and try again!\n";
  print "Plotting this file might cause damage to your device!!!!\n";
  print "########################################################\n\n";
}
if(($y_max - $y_min) > $bed_size_y){
  print "\n\n########################################################\n";
  print "Height of plot larger than height of print bed !!!!\n";
  print "Change value of \$scale and try again!\n";
  print "Plotting this file might cause damage to your device!!!!\n";
  print "########################################################\n\n";
}



#++++++++++++++++++++ Add set of coordinates +++++++++++++++++++++++++++
sub AddCoordinates{
  $coord_no++;
  push @remarks, "";
  push @feedrate_array, $feed_rate;
  push @X_coords, $X;
  push @Y_coords, $Y;
  push @Z_coords, $Z;

  if($X > $x_max){
     $x_max = $X;
  }
  if($X < $x_min){
    $x_min = $X;
  }
  if($Y > $y_max){
    $y_max = $Y;
  }
  if($Y < $y_min){
    $y_min = $Y;
  }


}

#+++++++++++++++++++++++++ Park Pen ++++++++++++++++++++++++++++++++++++
sub ParkPen{

  print"Parking pen...\n";

  $coord_no++;
  push @remarks, "Parking pen $_[0] (keep X keep Y). Going to low pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, -999999.999999;
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...X pos gap...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...high pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...X pos parking...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Parking pen $_[0] (keep Y). ...low pos parking done!";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];

  print"...parked!\n";
}


#+++++++++++++++++++++++++ Pick Pen ++++++++++++++++++++++++++++++++++++
sub PickPen{

  $coord_no++;
  push @remarks, "Picking pen $_[0] (keep X keep Y). Going to low pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, -999999.999999;
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low_empty[$_[0]];

  my $amplitude = 0.5;
  my $i;
  my $sign = 0;
  for($i=$pen_Z_low_empty[$_[0]]; $i<$pen_Z_high[$_[0]]; $i+=$amplitude*2){
    $coord_no++;
    push @remarks, "PenAction: Pen pen $_[0] (keep Y). ...shaking up...";
    push @feedrate_array, $feed_rate_pen_change;
    if($sign==0){
      push @X_coords, $pen_X[$_[0]]+$amplitude;
      $sign=1;
    }
    else{
      push @X_coords, $pen_X[$_[0]]-$amplitude;
      $sign=0;
    }
    push @Y_coords, -999999.999999;
    push @Z_coords, $i;
  }

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...high pos...";
  push @feedrate_array, $feed_rate_pen_change;

  push @X_coords, $pen_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos gap high...";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_high[$_[0]];

  $coord_no++;
  push @remarks, "PenAction: Picking pen $_[0] (keep Y). ...X pos gap low, done!";
  push @feedrate_array, $feed_rate_pen_change;
  push @X_coords, $pen_gap_X[$_[0]];
  push @Y_coords, -999999.999999;
  push @Z_coords, $pen_Z_low[$_[0]];

}
